// HelixStringFragmentation.h is a part of the PYTHIA event generator, adapted by S.Todorova.
// Copyright (C) 2012 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This file contains the classes for string fragmentation.
// HelixStringEnd: keeps track of the fragmentation step.
// HelixStringFragmentation: is the top-level class.

#ifndef Pythia8_HelixStringFragmentation_H
#define Pythia8_HelixStringFragmentation_H

#include "Basics.h"
#include "Event.h"
#include "Info.h"
#include "FragmentationFlavZpT.h"
#include "HelixFragmentationSystems.h"
#include "ParticleData.h"
#include "PythiaStdlib.h"
#include "Settings.h"

namespace Pythia8 {
 
//==========================================================================

// The HelixStringEnd class contains the information related to 
// one of the current endpoints of the string system.
// Only to be used inside HelixStringFragmentation, so no private members.

class HelixStringEnd {

public:

  // Constructor. 
  HelixStringEnd() { }

  // Save pointers.
  void init( Settings& settings, ParticleData* particleDataPtrIn, StringFlav* flavSelPtrIn, 
	     StringZ* zSelPtrIn, Rndm* rndmPtrIn);
   
  // Set up initial endpoint values from input.
  void setUp(bool fromPosIn, int iEndIn, int idOldIn, int iMaxIn,
    double pxIn, double pyIn, double GammaIn, double xPosIn, double xNegIn); 

  // Fragment off one hadron from the string system, in flavour and pT.
  void newHadron(int hadMode=0);

  // Fragment off one hadron from the string system, in momentum space,
  // by taking steps either from positive or from negative end.
  Vec4 kinematicsHadron(HelixStringSystem& system, HadronSystem& hadron);

  // Update string end information after a hadron has been removed.
  void update();

  // calculate helix phase difference 
  double phaseDifference(vector<pair<int,double> > pos,vector<pair<int,double> > neg,  HelixStringSystem& system);
  double phaseDifference(pair<int,double> pos, pair<int,double>  neg,  HelixStringSystem& system);

  // Constants: could only be changed in the code itself.
  static const double TINY, PT2SAME;

  // Pointer to the particle data table.
  ParticleData* particleDataPtr;

  // Pointers to classes for flavour, pT and z generation.
  StringFlav*   flavSelPtr;
  //StringPT*     pTSelPtr;
  StringZ*      zSelPtr;
  Rndm*         rndmPtr;
 
  // Data members.
  bool   fromPos;
  int    iEnd, iMax, idHad, iPosOld, iNegOld, iPosNew, iNegNew;
  double pxOld, pyOld, pxNew, pyNew, pxHad, pyHad, mHad, mT2Had, zHad,
         GammaOld, GammaNew, xPosOld, xPosNew, xPosHad, xNegOld, xNegNew,
         xNegHad;
  double phaseOld, phaseNew;
  int    chirality; 
  FlavContainer flavOld, flavNew;
  Vec4   pHad, pSoFar;
  double radius, sigma_radius, screwiness;
};
  
//==========================================================================

// The HelixStringFragmentation class contains the top-level routines 
// to fragment a colour singlet partonic system.

class HelixStringFragmentation {

public:

  // Constructor. 
  HelixStringFragmentation() {}

  // Initialize and save pointers.
  void init(Info* infoPtrIn, Settings& settings, 
    ParticleData* particleDataPtrIn, Rndm* rndmPtrIn, 
    StringFlav* flavSelPtrIn, StringZ* zSelPtrIn);

  // Do the fragmentation: driver routine.
  bool fragment( int iSub, ColConfig& colConfig, Event& event);

  // Find the boost matrix to the rest frame of a junction.
  RotBstMatrix junctionRestFrame(Vec4& p0, Vec4& p1, Vec4& p2);

private: 

  // Constants: could only be changed in the code itself.
  static const int    NTRYFLAV, NTRYJOIN, NSTOPMASS, NTRYJNREST, 
                      NTRYJNMATCH, NTRYJRFEQ;
  static const double FACSTOPMASS, CLOSEDM2MAX, CLOSEDM2FRAC, EXPMAX,
                      MATCHPOSNEG, EJNWEIGHTMAX, CONVJNREST, M2MAXJRF, 
                      CONVJRFEQ;

  // Pointer to various information on the generation.
  Info*         infoPtr;

  // Pointer to the particle data table.
  ParticleData* particleDataPtr;

  // Pointer to the random number generator.
  Rndm*         rndmPtr;

  // Pointers to classes for flavour, pT and z generation.
  StringFlav*   flavSelPtr;
  //StringPT*     pTSelPtr;
  StringZ*      zSelPtr;

  // Initialization data, read from Settings.
  double stopMass, stopNewFlav, stopSmear, eNormJunction,
         eBothLeftJunction, eMaxLeftJunction, eMinLeftJunction, bLund;

  // Data members.
  bool   hasJunction, isClosed;
  int    iPos, iNeg;
  double w2Rem, stopMassNow;
  Vec4   pSum, pRem, pJunctionHadrons;
  double radius, sigma_radius, screwiness;         // helix parameters
  int    hadronMode;                 // temporary: forced direct hadrons
  bool   doHelixDecay; 

  // List of partons in string system.
  vector<int> iParton;

  // Temporary event record for the produced particles.
  Event hadrons;

  // Array saving parton content of direct hadrons for the decay
  vector<pair<int,HadronSystem> > resonances;

  // Information on the system of string regions.
  HelixStringSystem system, systemMin, systemMid;

  // Information on the two current endpoints of the fragmenting system.
  HelixStringEnd posEnd, negEnd; 

  // Find region where to put first string break for closed gluon loop.
  vector<int> findFirstRegion(vector<int>& iPartonIn, Event& event);

  // Set flavours and momentum position for initial string endpoints. 
  void setStartEnds(int idPos, int idNeg, HelixStringSystem systemNow);

  // get phase from pT components
  double  getPhase(double px, double py); 

  // Check remaining energy-momentum whether it is OK to continue.
  bool energyUsedUp(bool fromPos);

  // Produce the final two partons to complete the system.
  bool finalTwo(bool fromPos);

  // Construct a special joining region for the final two hadrons.
  StringRegion finalRegion();

  // Store the hadrons in the normal event record, ordered from one end.
  void store(Event& event);

  // Collect remaining momentum
  HadronSystem findRem();

  // helix decay
  bool splitHelix(HadronSystem&, int, int, int, int);

  // resonance decay ?
  bool helixDecay(int idHad);
  bool helixDecayResolved(int idHad,int&, int&, int&);
  void decayHelix();

  // Fragment off two of the string legs in to a junction. 
  bool fragmentToJunction(Event& event);

};  
 
//==========================================================================

} // end namespace Pythia8

#endif // Pythia8_HelixStringFragmentation_H
