// HelixStringFragmentation.cc is a part of the PYTHIA event generator, adapted by S.Todorova.
// Copyright (C) 2012 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// Function definitions (not found in the header) for the HelixStringEnd and
// HelixStringFragmentation classes.

#include "HelixStringFragmentation.h"

namespace Pythia8 {
 
//==========================================================================

// The HelixStringEnd class.

//--------------------------------------------------------------------------
 
// Constants: could be changed here if desired, but normally should not.

// Avoid unphysical solutions to equation system.
const double HelixStringEnd::TINY = 1e-6;

// Assume two (eX, eY) regions are related if pT2 differs by less.
const double HelixStringEnd::PT2SAME = 0.01;

//--------------------------------------------------------------------------
  void HelixStringEnd::init( Settings& settings, ParticleData* particleDataPtrIn, 
             StringFlav* flavSelPtrIn, StringZ* zSelPtrIn, Rndm* rndmPtrIn) {
    particleDataPtr = particleDataPtrIn; flavSelPtr = flavSelPtrIn;
    zSelPtr = zSelPtrIn;
    rndmPtr = rndmPtrIn;
    radius  = settings.parm("HelixStringFragmentation:helixRadius");
    sigma_radius  = settings.parm("HelixStringFragmentation:sigmaHelixRadius");
    screwiness  = settings.parm("HelixStringFragmentation:screwiness");    
  }

// Set up initial endpoint values from input.

void HelixStringEnd::setUp(bool fromPosIn, int iEndIn, int idOldIn, int iMaxIn,
  double pxIn, double pyIn, double GammaIn, double xPosIn, double xNegIn) {

  // Simple transcription from input.
  fromPos  = fromPosIn;
  iEnd     = iEndIn;
  iMax     = iMaxIn;
  flavOld  = FlavContainer(idOldIn);
  pxOld    = pxIn;
  pyOld    = pyIn;
  GammaOld = GammaIn;
  iPosOld  = (fromPos) ? 0 : iMax;
  iNegOld  = (fromPos) ? iMax : 0;
  xPosOld  = xPosIn;
  xNegOld  = xNegIn;

}

//--------------------------------------------------------------------------

// Fragment off one hadron from the string system, in flavour and pT.

void HelixStringEnd::newHadron(int hadMode) {

  if (abs(flavOld.id)<3 && hadMode==1) {        // force production of charged pions
    int idn = 3 - abs(flavOld.id);
    if ( flavOld.id > 0 ) idn *= -1;    
    flavNew.id = idn;
    idHad = (flavOld.id + flavNew.id)>0 ? 211 : -211;
  } else if (abs(flavOld.id)<3 && hadMode==2) {   // force production of rho0
    flavNew.id = -flavOld.id;
    idHad = 113;
  } else if (abs(flavOld.id)<4 && hadMode==3) {   // force production of direct pi/K
    if (abs(flavOld.id)==3) {       
      flavNew = flavSelPtr->pick( flavOld);
      while (abs(flavNew.id)>2) flavNew = flavSelPtr->pick( flavOld);
      if (abs(flavNew.id)==1) idHad = 311;
      else idHad = flavOld.id<0 ? 321 : -321;             
    } else {
      flavNew = flavSelPtr->pick( flavOld);
      while (abs(flavNew.id)>3) flavNew = flavSelPtr->pick( flavOld);
      if (abs(flavNew.id)==3) {
        if (abs(flavOld.id)==1) idHad = 311;
	else idHad = flavNew.id<0 ? 321 : -321;             
      } else {
        if ( (flavOld.id+flavNew.id)==0 ) idHad = 111;
        else idHad = (flavOld.id+flavNew.id) > 0 ? 211 : -211;  
      }
    }
  } else if (abs(flavOld.id)<4 && hadMode==4) {   // force production of direct rho/K*
    if (abs(flavOld.id)==3) {       
      flavNew = flavSelPtr->pick( flavOld);
      while (abs(flavNew.id)>2) flavNew = flavSelPtr->pick( flavOld);
      if (abs(flavNew.id)==1) idHad = 313;
      else idHad = flavOld.id<0 ? 323 : -323;             
    } else {
      flavNew = flavSelPtr->pick( flavOld);
      while (abs(flavNew.id)>3) flavNew = flavSelPtr->pick( flavOld);
      if (abs(flavNew.id)==3) {
	if (abs(flavOld.id)==1) idHad = 313;
	else idHad = flavNew.id<0 ? 323 : -323;             
      } else {
	if ( (flavOld.id+flavNew.id)==0 ) idHad = 113;
	else idHad = (flavOld.id+flavNew.id) > 0 ? 213 : -213;  
      }
    } 
  } else {
    // Pick new flavour and form a new hadron.
    do {
      flavNew = flavSelPtr->pick( flavOld);
      idHad   = flavSelPtr->combine( flavOld, flavNew);
    } while (idHad == 0);
  }

  // Pick its transverse momentum - will be recalculated later
  double phi = rndmPtr->flat()*2*M_PI;
  double rad = radius+sigma_radius*rndmPtr->gauss();
  pxNew = rad*cos(phi);
  pyNew = rad*sin(phi);

  pxHad = pxOld + pxNew;
  pyHad = pyOld + pyNew;

  // Pick its mass and thereby define its transverse mass.
  mHad   = particleDataPtr->mass(idHad);
  mT2Had = pow2(mHad) + pow2(pxHad) + pow2(pyHad);

}

//--------------------------------------------------------------------------

// Fragment off one hadron from the string system, in momentum space,
// by taking steps from positive end.

Vec4 HelixStringEnd::kinematicsHadron( HelixStringSystem& system, HadronSystem& hadron) {

  // Pick fragmentation step z and calculate new Gamma.
  zHad = zSelPtr->zFrag( flavOld.id, flavNew.id, mT2Had);
  //zHad = zSelPtr->zFrag( flavOld.id, flavNew.id, mHad*mHad);
  GammaNew = (1. - zHad) * (GammaOld + mT2Had / zHad); 
  
  // Set up references that are direction-neutral;
  // ...Dir for direction of iteration and ...Inv for its inverse.
  int&    iDirOld = (fromPos) ? iPosOld : iNegOld;
  int&    iInvOld = (fromPos) ? iNegOld : iPosOld;
  int&    iDirNew = (fromPos) ? iPosNew : iNegNew;
  int&    iInvNew = (fromPos) ? iNegNew : iPosNew;
  double& xDirOld = (fromPos) ? xPosOld : xNegOld; 
  double& xInvOld = (fromPos) ? xNegOld : xPosOld; 
  double& xDirNew = (fromPos) ? xPosNew : xNegNew; 
  double& xInvNew = (fromPos) ? xNegNew : xPosNew; 
  double& xDirHad = (fromPos) ? xPosHad : xNegHad; 
  double& xInvHad = (fromPos) ? xNegHad : xPosHad; 

  // Start search for new breakup in the old region.
  iDirNew = iDirOld;
  iInvNew = iInvOld;
  Vec4 pTNew;

  // vector holding trace of composition of the hadron momentum
  vector< pair < int,double > > hPos; 
  vector< pair < int, double > > hNeg;
  pair<Vec4,Vec4> oPt;  

  // Each step corresponds to trying a new string region.
  for (int iStep = 0; ; ++iStep) { 

    // Referance to current string region.
    StringRegion& region = system.region( iPosNew, iNegNew);

    // Now begin special section for rapid processing of low region.
    if (iStep == 0 && iPosOld + iNegOld == iMax) { 

      // A first step within a low region is easy.
      if (mT2Had < zHad * xDirOld * (1. - xInvOld) * region.w2) {

        // Translate into x coordinates.
        xDirHad = zHad * xDirOld;
        xInvHad = mT2Had / (xDirHad * region.w2);

        bool nextReg=false;
        // step in to adjust for helix phase
	// calculate phase difference from fractions of parton momenta
	hPos.push_back(std::pair<int,double> (iPosNew, xPosHad));
	hNeg.push_back(std::pair<int,double> (iNegNew, xNegHad));
	double phaseHad = phaseDifference(hPos,hNeg,system); 
	phaseNew = phaseOld + chirality*phaseHad; 
	
	double ptOld = sqrt(pxOld*pxOld +pyOld*pyOld);
	double ptNew = sqrt(pxNew*pxNew +pyNew*pyNew);
	
	if ( ptOld < TINY ) {            // trivial adjustment
	  pxNew = ptNew*cos(phaseNew-chirality*0.5*M_PI);
	  pyNew = ptNew*sin(phaseNew-chirality*0.5*M_PI);
	  pxHad = pxNew+pxOld;
	  pyHad = pyNew+pyOld;
	} else {
	  // recalculate pT components
	  double pxn = ptNew*cos(phaseNew-chirality*0.5*M_PI);
	  double pyn = ptNew*sin(phaseNew-chirality*0.5*M_PI);
	  double pxHadn = pxOld + pxn;
	  double pyHadn = pyOld + pyn;
	  double pT2n = pxHadn*pxHadn + pyHadn*pyHadn;
	  double mT2n = pT2n + pow2(mHad);
          
	  if (mT2n > zHad * xDirOld * (1. - xInvOld) * region.w2) nextReg=true;
	  if (!nextReg ) {      // there should be a solution in this region
	    // keep phase difference constant
	    double ph2 = phaseDifference(pair<int,double>(iPosNew,xPosHad),pair<int,double>(iNegNew,xNegHad), system);    
	    double mp = 0.5*sqrt(system.regionLowPos(iPosNew).w2);
	    double mn = 0.5*sqrt(system.regionLowNeg(iNegNew).w2);
	    //// recalculate fractions 
	    double d2 = ph2*ph2/mp/mp-4*mT2n/region.w2*mn/mp;
	    if (d2 < 0.) return Vec4(0., 0., 0., -1.); 
	    double sol1 = 0.5*ph2/mp+0.5*sqrt(d2);
	    double xsol1 = mT2n / (sol1 * region.w2);
	    double sol2 = 0.5*ph2/mp-0.5*sqrt(d2); 
	    double xsol2 = mT2n / (sol2 * region.w2);
            
	    // pick the closest solution
	    if ( abs(sol1-xPosHad)+abs(xsol1-xNegHad) <  abs(sol2-xPosHad)+abs(xsol2-xNegHad) ) {
	      xPosHad = sol1; 
	      xNegHad = xsol1; 
	    } else {
	      xPosHad = sol2; 
	      xNegHad = xsol2; 
	    }
            
	    pxNew = pxn;
	    pyNew = pyn;
	    pxHad = pxHadn;
	    pyHad = pyHadn;
	    mT2Had = mT2n;
	  }	    
	}

	if (!nextReg) {
	  xDirNew = xDirOld - xDirHad;
	  xInvNew = xInvOld + xInvHad;
	  
	  // Find and return four-momentum of the produced particle.
          if (hPos.size()>0) {
	    hPos.back() = std::pair<int,double> (iPosNew, xPosHad);
	    hNeg.back() = std::pair<int,double> (iNegNew, xNegHad);
	  } else {
	    hPos.push_back(std::pair<int,double> (iPosNew, xPosHad));
	    hNeg.push_back(std::pair<int,double> (iNegNew, xNegHad));
	  }
	  oPt=std::pair<Vec4,Vec4> (pxHad*region.eX, pyHad*region.eY);
	 
	  hadron.frPos.push_back(pair<int,double>(iPosNew,xPosHad));
	  hadron.frNeg.push_back(pair<int,double>(iNegNew,xNegHad));
          if (fromPos) hadron.setEnds(iPosOld,iNegOld,pxOld,pyOld,iPosNew,iNegNew,pxHad-pxOld,pyHad-pyOld);
          else  hadron.setEnds(iPosNew,iNegNew,pxHad-pxOld,pyHad-pyOld,iPosOld,iNegOld,pxOld,pyOld);
          if (fromPos) hadron.setPhases(phaseOld,phaseNew);
          else  hadron.setPhases(phaseNew,phaseOld);

          system.fractions[iPosNew].first -=xPosHad; 
          system.fractions[iNegNew].second -=xNegHad;
  
	  return region.pHad( xPosHad, xNegHad, pxHad, pyHad);
	}
 
      // A first step out of a low region also OK, if there are more regions. 
      // Negative energy signals failure, i.e. in last region.
      } 

      int iPosCurr=iPosNew;
      int iNegCurr=iNegNew;
      
      --iInvNew; 
      if (iInvNew < 0) return Vec4(0., 0., 0., -1.); 
      
      // Momentum taken by stepping out of region. Continue to next region.
      xInvHad = 1. - xInvOld;
      xDirHad = 0.;  

      pSoFar  = region.pHad( xPosHad, xNegHad, pxOld, pyOld);

      hadron.frPos.push_back(pair<int,double>(iPosCurr,xPosHad));
      hadron.frNeg.push_back(pair<int,double>(iNegCurr,xNegHad));
      hadron.setEnds(iPosCurr,iNegCurr,pxOld,pyOld,0,0,0.,0.);
      system.fractions[iPosCurr].first -=xPosHad; 
      system.fractions[iNegCurr].second -=xNegHad; 
      hPos.push_back(std::pair<int,double> (iPosCurr, xPosHad));
      hNeg.push_back(std::pair<int,double> (iNegCurr, xNegHad));
      
      continue;
 
    // Else, for first step, take into account starting pT.
    } else if (iStep == 0) {
      pSoFar = region.pHad( 0., 0., pxOld, pyOld);
      pTNew  = region.pHad( 0., 0., pxNew, pyNew);
      oPt=std::pair<Vec4,Vec4> (pxOld*region.eX+pyOld*region.eY,0.*region.eX+0.*region.eY);
      hadron.setEnds(iPosNew,iNegNew,pxOld,pyOld,0,0,0.,0.);
    }

    // Now begin normal treatment of nontrivial regions.
    // Set up four-vectors in a region not visited before.
    if (!region.isSetUp) region.setUp( 
      system.regionLowPos(iPosNew).pPos,
      system.regionLowNeg(iNegNew).pNeg, true);

    // If new region is vanishingly small, continue immediately to next.
    // Negative energy signals failure to do this, i.e. moved too low.
    if (region.isEmpty) {
      //cout <<"empty region in:"<<hadron.getMomentum(system)<<","<<pSoFar << endl;
      xDirHad = (iDirNew == iDirOld) ? xDirOld : 1.;
      xInvHad = 0.;
      pSoFar += region.pHad( xPosHad, xNegHad, 0., 0.);
 
      ++iDirNew; 
      if (iDirNew + iInvNew > iMax) return Vec4(0., 0., 0., -1.); 
      continue;
    }

    // Reexpress pTNew w.r.t. base vectors in new region, if possible.
    // Recall minus sign from normalization e_x^2 = e_y^2 = -1.
    double pxNewTemp = -pTNew * region.eX;
    double pyNewTemp = -pTNew * region.eY;
    
    double a1=0.5*M_PI;
    if (abs(pxNew)>0.) a1=atan(pyNew/pxNew);
    if (pxNew<0.) a1+=M_PI;
  
    if (abs( pxNewTemp * pxNewTemp + pyNewTemp * pyNewTemp 
      - pxNew * pxNew - pyNew * pyNew) < PT2SAME) {
      pxNew = pxNewTemp;
      pyNew = pyNewTemp;
    }

    double a2=0.5*M_PI;
    if (abs(pxNew)>0.) a2=atan(pyNew/pxNew);
    if (pxNew<0.) a2+=M_PI;

    // Four-momentum taken so far, including new pT.
    Vec4 pTemp = pSoFar + region.pHad( 0., 0., pxNew, pyNew);

    // Derive coefficients for m2 expression.
    // cM2 * x+ + cM3 * x- + cM4 * x+ * x- = m^2 - cM1;
    double cM1 = pTemp.m2Calc();
    double cM2 = 2. * (pTemp * region.pPos); 
    double cM3 = 2. * (pTemp * region.pNeg); 
    double cM4 = region.w2;
    if (!fromPos) swap( cM2, cM3);

    // Derive coefficients for Gamma expression. 
    // cGam2 * x+ + cGam3 * x- + cGam4 * x+ * x- = Gamma_new - cGam1;
    double cGam1 = 0.;
    double cGam2 = 0.;
    double cGam3 = 0.;
    double cGam4 = 0.;
    for (int iInv = iInvNew; iInv <= iMax - iDirNew; ++iInv) {
      double xInv = 1.;
      if (iInv == iInvNew) xInv = (iInvNew == iInvOld) ? xInvOld : 0.;
      for (int iDir = iDirNew; iDir <= iMax - iInv; ++iDir) {
        double xDir = (iDir == iDirOld) ? xDirOld : 1.; 
        int iPos = (fromPos) ? iDir : iInv;
        int iNeg = (fromPos) ? iInv : iDir;
        StringRegion& regionGam =  system.region( iPos, iNeg);
        if (!regionGam.isSetUp) regionGam.setUp( 
          system.regionLowPos(iPos).pPos, 
          system.regionLowNeg(iNeg).pNeg, true);
        double w2 = regionGam.w2;
        cGam1 += xDir * xInv * w2;
        if (iDir == iDirNew) cGam2 -= xInv * w2;
        if (iInv == iInvNew) cGam3 += xDir * w2; 
        if (iDir == iDirNew && iInv == iInvNew) cGam4 -= w2;
      }
    }

    // Solve (m2, Gamma) equation system => r2 * x-^2 + r1 * x- + r0 = 0.
    double cM0   = pow2(mHad) - cM1;
    double cGam0 = GammaNew - cGam1;
    double r2    = cM3 * cGam4 - cM4 * cGam3;
    double r1    = cM4 * cGam0 - cM0 * cGam4 + cM3 * cGam2 - cM2 * cGam3;    
    double r0    = cM2 * cGam0 - cM0 * cGam2;
    double root  = sqrtpos( r1*r1 - 4. * r2 * r0 );
    if (abs(r2) < TINY || root < TINY) return Vec4(0., 0., 0., -1.); 
    xInvHad      = 0.5 * (root / abs(r2) - r1 / r2);
    xDirHad      = (cM0 - cM3 * xInvHad) / (cM2 + cM4 * xInvHad); 

    // Define position of new trial vertex.
    xDirNew = (iDirNew == iDirOld) ? xDirOld - xDirHad : 1. - xDirHad;
    xInvNew = (iInvNew == iInvOld) ? xInvOld + xInvHad : xInvHad;
 

    // calculate phase difference, adjust pT, recalculate splitting points
    hPos.push_back(pair<int,double> (iPosNew, xPosHad));
    hNeg.push_back(pair<int,double> (iNegNew, xNegHad));
    
    // calculate phase difference from fractions of parton momenta
    double phaseHad = phaseDifference(hPos,hNeg,system);
    phaseNew = phaseOld + chirality*phaseHad; 
    
    double ptNew = sqrt(pxNew*pxNew +pyNew*pyNew);
    double pxn = ptNew*cos(phaseNew-chirality*0.5*M_PI);
    double pyn = ptNew*sin(phaseNew-chirality*0.5*M_PI);
    
    pTemp = pSoFar + region.pHad( 0., 0., pxn, pyn);
    // cM2 * x+ + cM3 * x- + cM4 * x+ * x- = m^2 - cM1;
    cM1 = pTemp.m2Calc();
    cM2 = 2. * (pTemp * region.pPos); 
    cM3 = 2. * (pTemp * region.pNeg); 
    if (!fromPos) swap( cM2, cM3);
    cM0   = pow2(mHad) - cM1;
    //xDirHad      = (cM0 - cM3 * xInvHad) / (cM2 + cM4 * xInvHad); 
    xInvHad      = (cM0 - cM2 * xDirHad) / (cM3 + cM4 * xDirHad); 
    pxNew = pxn;
    pyNew = pyn;
    
    xDirNew = (iDirNew == iDirOld) ? xDirOld - xDirHad : 1. - xDirHad;
    xInvNew = (iInvNew == iInvOld) ? xInvOld + xInvHad : xInvHad;
    
    hPos.resize(hPos.size()-1);
    hNeg.resize(hNeg.size()-1);
    
    // Step up to new region if new x- > 1.
    if (xInvNew > 1.) {
      xInvHad = (iInvNew == iInvOld) ? 1. - xInvOld : 1.;
      xDirHad = 0.;
      pSoFar += region.pHad( xPosHad, xNegHad, 0., 0.);
      hPos.push_back(pair<int,double> (iPosNew, xPosHad));
      hNeg.push_back(pair<int,double> (iNegNew, xNegHad));

      hadron.frPos.push_back(pair<int,double>(iPosNew,xPosHad));
      hadron.frNeg.push_back(pair<int,double>(iNegNew,xNegHad));
      system.fractions[iPosNew].first -=xPosHad; 
      system.fractions[iNegNew].second -=xNegHad; 

      --iInvNew;
      if (iInvNew < 0) return Vec4(0., 0., 0., -1.); 
      continue;

    // Step down to new region if new x+ < 0.
    } else if (xDirNew < 0.) {
      xDirHad = (iDirNew == iDirOld) ? xDirOld : 1.; 
      xInvHad = 0.;      
      pSoFar += region.pHad( xPosHad, xNegHad, 0., 0.);
      hPos.push_back(pair<int,double> (iPosNew, xPosHad));
      hNeg.push_back(pair<int,double> (iNegNew, xNegHad));

      hadron.frPos.push_back(pair<int,double>(iPosNew,xPosHad));
      hadron.frNeg.push_back(pair<int,double>(iNegNew,xNegHad));
      system.fractions[iPosNew].first -=xPosHad; 
      system.fractions[iNegNew].second -=xNegHad; 

      ++iDirNew;
      if (iDirNew + iInvNew > iMax) return Vec4(0., 0., 0., -1.); 
      continue;
    }

    // Else we have found the correct region, and can return four-momentum.
    hPos.push_back(pair<int,double> (iPosNew, xPosHad));
    hNeg.push_back(pair<int,double> (iNegNew, xNegHad));
        
    // adjust for helix phase if required
    // calculate phase difference from fractions of parton momenta
    phaseHad = phaseDifference(hPos,hNeg,system);
    phaseNew = phaseOld + chirality*phaseHad; 

    ptNew = sqrt(pxNew*pxNew +pyNew*pyNew);
    pxn = ptNew*cos(phaseNew-chirality*0.5*M_PI);
    pyn = ptNew*sin(phaseNew-chirality*0.5*M_PI);

    double ph2 = phaseDifference(pair<int,double>(iPosNew,xPosHad),pair<int,double>(iNegNew,xNegHad), system); 
    double mp = sqrt(system.regionLowPos(iPosNew).w2);
    double mn = sqrt(system.regionLowNeg(iNegNew).w2);

    Vec4 bTemp = pSoFar + region.pHad( 0., 0. , pxn, pyn);
    double R = 2*ph2/mn; 
    double bm = bTemp*region.pNeg;
    double ca = mp/mn*region.w2;
    double cb = 2*mp/mn*bm-2*bTemp*region.pPos-R*region.w2;
    double cc = mHad*mHad-bTemp*bTemp-2*R*bm;
    
    double d2 = cb*cb - 4.*ca*cc;
    if (d2<0.) {
      return Vec4(0., 0., 0., -1.);
    }
    double sol1 = 0.5*(-cb+sqrt(d2))/ca;
    double sol2 = 0.5*(-cb-sqrt(d2))/ca; 
    double xsol1 = R-sol1*mp/mn;
    double xsol2 = R-sol2*mp/mn;
  
    if (abs(sol1-xPosHad)+abs(xsol1-xNegHad) < abs(sol2-xPosHad)+abs(xsol2-xNegHad)) {
      xPosHad = sol1;
      xNegHad = xsol1;
    } else {
      xPosHad = sol2;
      xNegHad = xsol2; 
    }       
    pxNew = pxn;
    pyNew = pyn;
    
    hPos.back()=pair<int,double> (iPosNew, xPosHad);
    hNeg.back()=pair<int,double> (iNegNew, xNegHad);
    
    hadron.frPos.push_back(pair<int,double>(iPosNew,xPosHad));
    hadron.frNeg.push_back(pair<int,double>(iNegNew,xNegHad));
    int iO=hadron.iPP; int jO=hadron.iPN; double xO = hadron.xP; double yO = hadron.yP;
    if (fromPos) { 
      hadron.setEnds(iO,jO,xO,yO,iPosNew,iNegNew,pxNew,pyNew);
      hadron.setPhases(phaseOld,phaseNew);
    } else {
      hadron.setEnds(iPosNew,iNegNew,pxNew,pyNew,iO,jO,xO,yO);
      hadron.setPhases(phaseNew,phaseOld);
    }

    system.fractions[iPosNew].first -=xPosHad; 
    system.fractions[iNegNew].second -=xNegHad; 

    return pSoFar + region.pHad( xPosHad, xNegHad, pxNew, pyNew);

  // End of "infinite" loop of stepping to new region.
  }

}

//--------------------------------------------------------------------------

  double HelixStringEnd::phaseDifference(vector<pair<int,double> > pos,vector<pair<int,double> > neg, HelixStringSystem& system) {

    //  [rad]
 
    double mReg=0.;
    double phaseH = 0.;
    for (unsigned int i=0; i<pos.size();i++ ) {
      mReg = system.regionLowPos(pos[i].first).w2;
      mReg = 0.5*sqrt(mReg); 
      phaseH += pos[i].second * mReg;
    }
    for (unsigned int i=0; i<neg.size();i++ ) {
      mReg = system.regionLowNeg(neg[i].first).w2;
      mReg = 0.5*sqrt(mReg); 
      phaseH += neg[i].second * mReg;
    }

    return screwiness * phaseH;

  }

//--------------------------------------------------------------------------

  double HelixStringEnd::phaseDifference(pair<int,double>  pos, pair<int,double>  neg, HelixStringSystem& system) {
    // [GeV]

    double mReg=0.;
    double phaseH = 0.;
    
    mReg = system.regionLowPos(pos.first).w2;
    mReg = 0.5*sqrt(mReg); 
    phaseH += pos.second * mReg;
    
    mReg = system.regionLowNeg(neg.first).w2;
    mReg = 0.5*sqrt(mReg); 
    phaseH += neg.second * mReg;

    return phaseH;

  }

//--------------------------------------------------------------------------

// Update string end information after a hadron has been removed.

void HelixStringEnd::update() {

  flavOld.anti(flavNew);
  iPosOld  = iPosNew;
  iNegOld  = iNegNew;
  pxOld    = -pxNew;
  pyOld    = -pyNew;
  GammaOld = GammaNew;
  xPosOld  = xPosNew;
  xNegOld  = xNegNew;
  phaseOld = phaseNew;
}
  
//==========================================================================

// The HelixStringFragmentation class.

//--------------------------------------------------------------------------

// Constants: could be changed here if desired, but normally should not.
// These are of technical nature, as described for each.

// Maximum number of tries to (flavour-, energy) join the two string ends.
const int    HelixStringFragmentation::NTRYFLAV     = 10; 
const int    HelixStringFragmentation::NTRYJOIN     = 50; 

// The last few times gradually increase the stop mass to make it easier.
const int    HelixStringFragmentation::NSTOPMASS    = 10; 
const double HelixStringFragmentation::FACSTOPMASS  = 1.05; 

// For closed string, pick a Gamma by taking a step with fictitious mass.
const double HelixStringFragmentation::CLOSEDM2MAX  = 25.; 
const double HelixStringFragmentation::CLOSEDM2FRAC = 0.1; 

// Do not allow too large argument to exp function.
const double HelixStringFragmentation::EXPMAX       = 50.;

// Matching criterion that p+ and p- not the same (can happen in gg loop).
const double HelixStringFragmentation::MATCHPOSNEG  = 1e-6;

// For pull on junction, do not trace too far down each leg.
const double HelixStringFragmentation::EJNWEIGHTMAX = 10.;

// Iterate junction rest frame boost until convergence or too many tries.
const double HelixStringFragmentation::CONVJNREST   = 1e-5;
const int    HelixStringFragmentation::NTRYJNREST   = 20; 

// Fail and try again when two legs combined to diquark (3 loops).
const int    HelixStringFragmentation::NTRYJNMATCH  = 20;

// Consider junction-leg parton as massless if m2 tiny.
const double HelixStringFragmentation::M2MAXJRF     = 1e-4;

// Iterate junction rest frame equation until convergence or too many tries. 
const double HelixStringFragmentation::CONVJRFEQ    = 1e-12;
const int    HelixStringFragmentation::NTRYJRFEQ    = 40; 

//--------------------------------------------------------------------------

// Initialize and save pointers. 

void HelixStringFragmentation::init(Info* infoPtrIn, Settings& settings, 
  ParticleData* particleDataPtrIn, Rndm* rndmPtrIn, StringFlav* flavSelPtrIn,  
  StringZ* zSelPtrIn) {

  // Save pointers.
  infoPtr         = infoPtrIn;
  particleDataPtr = particleDataPtrIn;
  rndmPtr         = rndmPtrIn;
  flavSelPtr      = flavSelPtrIn;
  //pTSelPtr        = pTSelPtrIn;
  zSelPtr         = zSelPtrIn;

  // Initialize the HelixStringFragmentation class.
  stopMass        = zSelPtr->stopMass();
  stopNewFlav     = zSelPtr->stopNewFlav();
  stopSmear       = zSelPtr->stopSmear();
  eNormJunction   = settings.parm("StringFragmentation:eNormJunction");
  eBothLeftJunction 
     = settings.parm("StringFragmentation:eBothLeftJunction");
  eMaxLeftJunction 
    = settings.parm("StringFragmentation:eMaxLeftJunction");
  eMinLeftJunction 
    = settings.parm("StringFragmentation:eMinLeftJunction");

  // Initialize the b parameter of the z spectrum, used when joining jets.
  bLund           = zSelPtr->bAreaLund();

  // Initialize the hadrons instance of an event record.
  hadrons.init( "(string fragmentation)", particleDataPtr);

  // Initialize the helix structure parameters.
  radius  = settings.parm("HelixStringFragmentation:helixRadius");
  sigma_radius  = settings.parm("HelixStringFragmentation:sigmaHelixRadius");
  screwiness  = settings.parm("HelixStringFragmentation:screwiness");    
  hadronMode  = settings.mode("HelixStringFragmentation:HadronMode");    
  doHelixDecay  = settings.flag("HelixStringFragmentation:helixDecay");    

  // Send on pointers to the two HelixStringEnd instances.
  posEnd.init( settings, particleDataPtr, flavSelPtr, zSelPtr, rndmPtr);   
  negEnd.init( settings, particleDataPtr, flavSelPtr, zSelPtr, rndmPtr);   

}

//--------------------------------------------------------------------------

// Perform the fragmentation.

bool HelixStringFragmentation::fragment( int iSub, ColConfig& colConfig, 
  Event& event) {

  // Find partons and their total four-momentum.
  iParton            = colConfig[iSub].iParton;
  iPos               = iParton[0];
  if (iPos < 0) iPos = iParton[1];
  int idPos          = event[iPos].id();
  iNeg               = iParton.back();
  int idNeg          = event[iNeg].id();
  pSum               = colConfig[iSub].pSum;

  // Reset the local event record.
  hadrons.clear();
  if (doHelixDecay) resonances.clear();
 
  // For closed gluon string: pick first breakup region.
  isClosed = colConfig[iSub].isClosed;
  if (isClosed) iParton = findFirstRegion(iParton, event);

  // For junction topology: fragment off two of the string legs. 
  // Then iParton overwritten to remaining leg + leftover diquark.
  pJunctionHadrons = 0.;
  hasJunction = colConfig[iSub].hasJunction;
  if (hasJunction && !fragmentToJunction(event)) return false;
  int junctionHadrons = hadrons.size(); 
  if (hasJunction) { 
    idPos  = event[ iParton[0] ].id();
    idNeg  = event.back().id();
    pSum  -= pJunctionHadrons;
  }

  // Set up kinematics of string evolution ( = motion).
  system.setUp(iParton, event);
  stopMassNow = stopMass;

  // Fallback loop, when joining in the middle fails.  Bailout if stuck.
  for ( int iTry = 0; ; ++iTry) {
    if (iTry > NTRYJOIN) {
      infoPtr->errorMsg("Error in HelixStringFragmentation::fragment: " 
        "stuck in joining");
      if (hasJunction) event.popBack(1);
      return false;
    } 

    // After several failed tries gradually allow larger stop mass.
    if (iTry > NTRYJOIN - NSTOPMASS) stopMassNow *= FACSTOPMASS;

    // Set up flavours of two string ends, and reset other info.
    setStartEnds(idPos, idNeg, system);
    pRem = pSum;
    system.resetFractions();

    // Begin fragmentation loop, interleaved from the two ends.
    bool fromPos;
    Vec4 psumHad;
    int idec = 0;
    bool fromPosDec=false;
    bool fromNegDec=false;
    bool dec = false;
    bool notEnd = false;

    for ( ; ; ) {

      // Take a step either from the positive or the negative end.
      fromPos           = (rndmPtr->flat() < 0.5);
      HelixStringEnd& nowEnd = (fromPos) ? posEnd : negEnd;
     
      // Construct trial hadron and check that energy remains.
      nowEnd.newHadron(hadronMode);
      if ( energyUsedUp(fromPos) ) break;

      // Save hadron string composition for decays
      HadronSystem hadron;

      // backup in case of failed decay
      system.saveFractions();  

      // Construct kinematics of the new hadron and store it.
      Vec4 pHad = nowEnd.kinematicsHadron(system, hadron);
      int statusHad = (fromPos) ? 83 : 84; 
      hadrons.append( nowEnd.idHad, statusHad, iPos, iNeg, 
        0, 0, 0, 0, pHad, nowEnd.mHad);
      if (pHad.e() < 0.) break;

      Vec4 hadm = hadron.getMomentum(system);
      psumHad += hadm; 

      // Helix string driven decay
      dec = true;         

       
      // use arguments to pass flavour combination and return daughter id's
      int id1 = fromPos ? nowEnd.flavOld.id : nowEnd.flavNew.id; 
      int id2 = fromPos ? nowEnd.flavNew.id : nowEnd.flavOld.id;
      int idnext = fromPos ? 1 : 0 ;

      notEnd = (fromPos && fromPosDec) || (!fromPos && fromNegDec);

      if (doHelixDecay && notEnd && helixDecayResolved(nowEnd.idHad,id1,id2,idnext) ) {
     
        dec = splitHelix(hadron,id1,id2,hadrons.size()-1,idnext) ; 
      }
      if (!dec && idec<10) { idec++; system.restoreFractions(); hadrons.popBack(1); continue;}      
      idec = 0; 

      // Update string end and remaining momentum.
      nowEnd.update();
      pRem -= pHad;
      if (fromPos) fromPosDec = true;
      else fromNegDec = true;

    // End of fragmentation loop.
    }
   
    // When done, join in the middle. If this works, then really done.
    if ( finalTwo(fromPos) ) break;

    // Else remove produced particles (except from first two junction legs)
    // and start all over.
    int newHadrons = hadrons.size() - junctionHadrons;
    hadrons.popBack(newHadrons);
  }

  // Junctions: remove fictitious end, restore original parton list
  if (hasJunction) {
    event.popBack(1);
    iParton = colConfig[iSub].iParton;
  }

  // decay according to the helix structure
  if (doHelixDecay && resonances.size()) decayHelix();

  // Store the hadrons in the normal event record, ordered from one end.
  store(event);

  // Done.
  return true;

}

//--------------------------------------------------------------------------

// Find region where to put first string break for closed gluon loop.
  
vector<int> HelixStringFragmentation::findFirstRegion(vector<int>& iPartonIn,
  Event& event) {

  // Evaluate mass-squared for all adjacent gluon pairs.
  vector<double> m2Pair; 
  double m2Sum = 0.;
  int size = iPartonIn.size();
  for (int i = 0; i < size; ++i) {
    double m2Now = 0.5 * event[ iPartonIn[i] ].p() 
      * event[ iPartonIn[(i + 1)%size] ].p();  
    m2Pair.push_back(m2Now);
    m2Sum += m2Now;
  }
   
  // Pick breakup region with probability proportional to mass-squared.
  double m2Reg = m2Sum * rndmPtr->flat();
  int iReg = -1;
  do m2Reg -= m2Pair[++iReg];
  while (m2Reg > 0. && iReg < size - 1); 

  // Create reordered parton list, with breakup string region duplicated.
  vector<int> iPartonOut;   
  for (int i = 0; i < size + 2; ++i) 
    iPartonOut.push_back( iPartonIn[(i + iReg)%size] );    

  // Done.
  return iPartonOut;
 
}

//--------------------------------------------------------------------------

// Set flavours and momentum position for initial string endpoints. 

void HelixStringFragmentation::setStartEnds( int idPos, int idNeg,
  HelixStringSystem systemNow) {

  // Variables characterizing string endpoints: defaults for open string.
  double px          = 0.;
  double py          = 0.;
  double Gamma       = 0.;
  double xPosFromPos = 1.;
  double xNegFromPos = 0.;
  double xPosFromNeg = 0.;
  double xNegFromNeg = 1.;

  // For closed gluon loop need to pick an initial flavour.
  if (isClosed) {
    do {
      int idTry = flavSelPtr->pickLightQ();
      FlavContainer flavTry(idTry, 1);
      flavTry = flavSelPtr->pick( flavTry);
      flavTry = flavSelPtr->pick( flavTry);
      idPos   = flavTry.id;
      idNeg   = -idPos;
    } while (idPos == 0);

    // Also need pT and breakup vertex position in region.
    double phi = rndmPtr->flat()*2*M_PI;
    double rad = radius+sigma_radius*rndmPtr->gauss();
    px = rad*cos(phi);
    py = rad*sin(phi);
    double m2Region = systemNow.regionLowPos(0).w2;
    double m2Temp   = min( CLOSEDM2MAX, CLOSEDM2FRAC * m2Region);
    do {
      double zTemp = zSelPtr->zFrag( idPos, idNeg, m2Temp);
      xPosFromPos  = 1. - zTemp;
      xNegFromPos  = m2Temp / (zTemp * m2Region);
    } while (xNegFromPos > 1.);
    Gamma = xPosFromPos * xNegFromPos * m2Region;
    xPosFromNeg = xPosFromPos;
    xNegFromNeg = xNegFromPos; 
  }

  // Initialize two string endpoints.
  posEnd.setUp(  true, iPos, idPos, systemNow.iMax,  px,  py, 
    Gamma, xPosFromPos, xNegFromPos);
  negEnd.setUp( false, iNeg, idNeg, systemNow.iMax, -px, -py, 
    Gamma, xPosFromNeg, xNegFromNeg); 

  // For closed gluon loop can allow popcorn on one side but not both.
  if (isClosed) {
    flavSelPtr->assignPopQ(posEnd.flavOld);
    flavSelPtr->assignPopQ(negEnd.flavOld);
    if (rndmPtr->flat() < 0.5) posEnd.flavOld.nPop = 0;    
    else                    negEnd.flavOld.nPop = 0;
    posEnd.flavOld.rank = 1;
    negEnd.flavOld.rank = 1;
  } 

  // phase info
  posEnd.phaseOld = rndmPtr->flat() * 2 * M_PI;  
  posEnd.chirality = (rndmPtr->flat()<0.5) ?  1 : -1;
  negEnd.phaseOld = posEnd.phaseOld + screwiness*posEnd.chirality*systemNow.phaseTot;
  negEnd.chirality = - posEnd.chirality;

  if (isClosed) {
    posEnd.phaseOld = getPhase(px,py);
    negEnd.phaseOld = posEnd.phaseOld;   // impose 
  }

  // Done.

}
//--------------------------------------------------------------------------

  double HelixStringFragmentation::getPhase(double px, double py) {

    double ph = 0.5*M_PI;
    if (abs(px) > 0.) ph = atan(py/px);
    if (px < 0. ) ph += M_PI;
 
    return ph; 

  }

//--------------------------------------------------------------------------

// Check remaining energy-momentum whether it is OK to continue.

bool HelixStringFragmentation::energyUsedUp(bool fromPos) {

  // If remaining negative energy then abort right away.
  if (pRem.e() < 0.) return true;

  // Calculate W2_minimum and done if remaining W2 is below it.
  double wMin = stopMassNow 
    + particleDataPtr->constituentMass(posEnd.flavOld.id)
    + particleDataPtr->constituentMass(negEnd.flavOld.id);
  if (fromPos) wMin += stopNewFlav 
    * particleDataPtr->constituentMass(posEnd.flavNew.id);
  else         wMin += stopNewFlav 
    * particleDataPtr->constituentMass(negEnd.flavNew.id);
  wMin *= 1. + (2. * rndmPtr->flat() - 1.) * stopSmear;
  w2Rem = pRem.m2Calc(); 
  if (w2Rem < pow2(wMin)) return true;

  // Else still enough energy left to continue iteration. 
  return false; 

}

//--------------------------------------------------------------------------

// Produce the final two partons to complete the system.

bool HelixStringFragmentation::finalTwo(bool fromPos) {
  
  // Check whether we went too far in p+-.
  if (pRem.e() < 0.  || w2Rem < 0. || (hadrons.size() > 0  
    && hadrons.back().e() < 0.) ) return false;  
  if ( posEnd.iPosOld > negEnd.iPosOld || negEnd.iNegOld > posEnd.iNegOld) 
    return false; 
  if ( posEnd.iPosOld == negEnd.iPosOld && posEnd.xPosOld < negEnd.xPosOld) 
    return false; 
  if ( posEnd.iNegOld == negEnd.iNegOld && posEnd.xNegOld > negEnd.xNegOld) 
    return false; 

  // Construct the final hadron from the leftover flavours. 
  // Impossible to join two diquarks. Also break if stuck for other reason.
  FlavContainer flav1 = (fromPos) ? posEnd.flavNew.anti() : posEnd.flavOld;
  FlavContainer flav2 = (fromPos) ? negEnd.flavOld : negEnd.flavNew.anti();
  if (flav1.isDiquark() && flav2.isDiquark()) return false;
  int idHad;
  for (int iTry = 0; iTry < NTRYFLAV; ++iTry) {
    idHad = flavSelPtr->combine( flav1, flav2);
    if (hadronMode>0) {
      if ( hadronMode<3) {
        if (abs(flav1.id)<3 && abs(flav2.id)<3) {
	  if (flav1.id+flav2.id == 1 ) idHad = hadronMode==1 ? 211 : 213;
	  else if (flav1.id+flav2.id == -1) idHad = hadronMode==1 ? -211 : -213;
	  else idHad = hadronMode==1 ? 111 : 113; 
	} 
      } else if (hadronMode<5) {
	if ( abs(flav1.id)<4 && abs(flav2.id)<4 ) {
	  if (abs(flav1.id)==3 && abs(flav2.id)==1) {
	    idHad = hadronMode==3 ? 311 : 313;
	  } else if (abs(flav1.id)==3 && abs(flav2.id)==2) {
	    if (flav1.id<0) idHad = hadronMode==3 ? 321 : 323;  
            else  idHad = hadronMode==3 ? -321 : -323;   
	  } else if (abs(flav1.id)==1 && abs(flav2.id)==3)  {
	    idHad = hadronMode==3 ? 311 : 313;
	  } else if (abs(flav1.id)==2 && abs(flav2.id)==3) {
	    if (flav1.id>0) idHad = hadronMode==3 ? 321 : 323;  
            else  idHad = hadronMode==3 ? -321 : -323;   
          } else if (flav1.id+flav2.id == 1 ) {
	    idHad = hadronMode==3 ? 211 : 213;
	  } else if (flav1.id+flav2.id == -1)  {
	    idHad = hadronMode==3 ? -211 : -213;
	  } else idHad = hadronMode==3 ? 111 : 113;
	} 
      }
    }
    if (idHad != 0) break;
  } 
  if (idHad == 0) return false;

  // Store the final particle and its new pT, and construct its mass.
  if (fromPos) {
    negEnd.idHad = idHad;
    negEnd.pxNew = -posEnd.pxNew;
    negEnd.pyNew = -posEnd.pyNew;
    negEnd.mHad  = particleDataPtr->mass(idHad);
  } else {     
    posEnd.idHad = idHad;
    posEnd.pxNew = -negEnd.pxNew;
    posEnd.pyNew = -negEnd.pyNew;
    posEnd.mHad  = particleDataPtr->mass(idHad);
  }

  HadronSystem remMom = findRem();
  remMom.setFlavours(posEnd.flavOld.id,negEnd.flavOld.id);
  int nextFlav = fromPos ? -flav1.id : flav2.id;  
  return splitHelix(remMom,posEnd.idHad,negEnd.idHad,0,nextFlav);
  
}
//--------------------------------------------------------------------------

//  Collect remaining momentum.

HadronSystem HelixStringFragmentation::findRem() {

  HadronSystem remP;
 
  for (unsigned int i=0; i<system.fractions.size(); i++)  {
    if (system.fractions[i].first!=0) remP.frPos.push_back(pair<int,double>(i,system.fractions[i].first));
    if (system.fractions[i].second!=0) remP.frNeg.push_back(pair<int,double>(i,system.fractions[i].second));
  }

  if (system.isGluonLoop) {
    remP.frPos.push_back(pair<int,double>(0,-1.));
    remP.frNeg.push_back(pair<int,double>(0,-1.));
  }

  // add pt
  remP.setEnds(posEnd.iPosOld,posEnd.iNegOld,posEnd.pxOld,posEnd.pyOld,
               negEnd.iPosOld,negEnd.iNegOld,negEnd.pxOld,negEnd.pyOld); 
 
  remP.setPhases(posEnd.phaseOld,negEnd.phaseOld);

  if ( abs((remP.getMomentum(system) - pRem).e())>0.1 ) {

   //cout << "problem with E/p conservation:"<<  remP.getMomentum(system) - pRem << endl;
   //for (unsigned int i=0; i<system.fractions.size(); i++)  {
   //  if (system.fractions[i].first!=0) cout<<i<<"+:"<<system.fractions[i].first*system.regionLowPos(i).pPos<<endl;
   //  if (system.fractions[i].second!=0) cout<<i<<"-:"<<system.fractions[i].second*system.regionLowNeg(i).pNeg<<endl;
   //}

  }

  return remP;
}


//--------------------------------------------------------------------------

// Store the hadrons in the normal event record, ordered from one end.

void HelixStringFragmentation::store(Event& event) {

  // Starting position.
  int iFirst = event.size();

  // Copy straight over from first two junction legs.
  if (hasJunction) { 
    for (int i = 0; i < hadrons.size(); ++i) 
    if (hadrons[i].status() == 85 || hadrons[i].status() == 86) 
      event.append( hadrons[i] );
  }
 
  // Loop downwards, copying all from the positive end.
  for (int i = 0; i < hadrons.size(); ++i) 
    if (hadrons[i].status() == 83) event.append( hadrons[i] );

  // Loop upwards, copying all from the negative end.
  for (int i = hadrons.size() - 1; i >= 0 ; --i) 
    if (hadrons[i].status() == 84) event.append( hadrons[i] );
  int iLast = event.size() - 1; 

  // collect daughters if decay
  for (int i = 0; i <= iLast; ++i)  {
    //cout <<event[i].status()<<","<<event[i].daughter1()<<","<<event[i].daughter2()<< endl;
    if (event[i].status()==83 || event[i].status()==84) {
      if (event[i].daughter1()>0) {
        hadrons[event[i].daughter1()].mother1(i);
        hadrons[event[i].daughter2()].mother1(i);
        hadrons[event[i].daughter1()].mother2(i);
        hadrons[event[i].daughter2()].mother2(i);
	event.append( hadrons[event[i].daughter1()] );
	event.append( hadrons[event[i].daughter2()] );
	event[i].statusNeg();
	event[i].daughter1(event.size()-2);
	event[i].daughter2(event.size()-1);
      }
    }
  }

  // Set decay vertex when this is displaced.
  if (event[posEnd.iEnd].hasVertex()) {
    Vec4 vDec = event[posEnd.iEnd].vDec();
    for (int i = iFirst; i <= iLast; ++i) event[i].vProd( vDec );
  }

  // Set lifetime of hadrons.
  for (int i = iFirst; i <= iLast; ++i) 
    event[i].tau( event[i].tau0() * rndmPtr->exp() );

  // Mark original partons as hadronized and set their daughter range.
  for (int i = 0; i < int(iParton.size()); ++i)
  if (iParton[i] >= 0) {
    event[ iParton[i] ].statusNeg();
    event[ iParton[i] ].daughters(iFirst, iLast);
  }    

}

//--------------------------------------------------------------------------

// Fragment off two of the string legs in to a junction. 

bool HelixStringFragmentation::fragmentToJunction(Event& event) {

  // Identify range of partons on the three legs.
  // (Each leg begins with an iParton[i] = -(10 + 10*junctionNumber + leg),
  // and partons then appear ordered from the junction outwards.)
  int legBeg[3] = { 0, 0, 0};
  int legEnd[3] = { 0, 0, 0};
  int leg = -1;
  // PS (4/10/2011) Protect against invalid systems 
  if (iParton[0] > 0) {
    infoPtr->errorMsg("Error in HelixStringFragmentation::fragment" 
		      "ToJunction: iParton[0] not a valid junctionNumber");
    return false;
  }
  for (int i = 0; i < int(iParton.size()); ++i) {
    if (iParton[i] < 0) {
      if (leg == 2) {
	infoPtr->errorMsg("Error in HelixStringFragmentation::fragment" 
			  "ToJunction: unprocessed multi-junction system");
	return false;
      }
      legBeg[++leg] = i + 1; 
    } 
    else legEnd[leg] = i;    
  }

  // Iterate from system rest frame towards the junction rest frame (JRF).
  RotBstMatrix MtoJRF, Mstep;
  MtoJRF.bstback(pSum);
  Vec4 pWTinJRF[3];
  int iter = 0;
  double errInCM = 0.;
  do { 
    ++iter;
  
    // Find weighted sum of momenta on the three sides of the junction.
    for (leg = 0; leg < 3; ++ leg) { 
      pWTinJRF[leg] = 0.; 
      double eWeight = 0.;
      for (int i = legBeg[leg]; i <= legEnd[leg]; ++i) {
        Vec4 pTemp = event[ iParton[i] ].p();
        pTemp.rotbst(MtoJRF);
        pWTinJRF[leg] += pTemp * exp(-eWeight);      
        eWeight += pTemp.e() / eNormJunction; 
        if (eWeight > EJNWEIGHTMAX) break; 
      }
    }

    // Store original deviation from 120 degree topology.
    if (iter == 1) errInCM = pow2(costheta(pWTinJRF[0], pWTinJRF[1]) + 0.5) 
      + pow2(costheta(pWTinJRF[0], pWTinJRF[2]) + 0.5) 
      + pow2(costheta(pWTinJRF[1], pWTinJRF[2]) + 0.5); 
   
    // Find new JRF from the set of weighted momenta.
    Mstep = junctionRestFrame( pWTinJRF[0], pWTinJRF[1], pWTinJRF[2]);
    // Fortran code will not take full step after the first few 
    // iterations. How implement this in terms of an M matrix??
    MtoJRF.rotbst( Mstep );
  } while (iter < 3 || (Mstep.deviation() > CONVJNREST && iter < NTRYJNREST) );

  // If final deviation from 120 degrees is bigger than in CM then revert.
  double errInJRF = pow2(costheta(pWTinJRF[0], pWTinJRF[1]) + 0.5) 
    + pow2(costheta(pWTinJRF[0], pWTinJRF[2]) + 0.5) 
    + pow2(costheta(pWTinJRF[1], pWTinJRF[2]) + 0.5); 
  if (errInJRF > errInCM) {
    infoPtr->errorMsg("Warning in HelixStringFragmentation::fragmentTo" 
      "Junction: bad convergence junction rest frame");
    MtoJRF.reset();
    MtoJRF.bstback(pSum); 
  } 

  // Opposite operation: boost from JRF to original system.
  RotBstMatrix MfromJRF = MtoJRF;
  MfromJRF.invert();

  // Sum leg four-momenta in original frame and in JRF.
  Vec4 pInLeg[3], pInJRF[3];
  for (leg = 0; leg < 3; ++leg) { 
    pInLeg[leg] = 0.; 
    for (int i = legBeg[leg]; i <= legEnd[leg]; ++i)  
      pInLeg[leg] += event[ iParton[i] ].p(); 
    pInJRF[leg] = pInLeg[leg]; 
    pInJRF[leg].rotbst(MtoJRF);
  }

  // Pick the two legs with lowest energy in JRF.
  int legMin = (pInJRF[0].e() < pInJRF[1].e()) ? 0 : 1;
  int legMax = 1 - legMin;
  if (pInJRF[2].e() < min(pInJRF[0].e(), pInJRF[1].e()) ) legMin = 2;
  else if (pInJRF[2].e() > max(pInJRF[0].e(), pInJRF[1].e()) ) legMax = 2;
  int legMid = 3 - legMin - legMax; 

  // Save info on which status codes belong with the three legs.
  int iJunction = (-iParton[0]) / 10 - 1;
  event.statusJunction( iJunction, legMin, 85);
  event.statusJunction( iJunction, legMid, 86);
  event.statusJunction( iJunction, legMax, 83); 

  // Temporarily copy the partons on the low-energy legs, into the JRF,
  // in reverse order, so (anti)quark leg end first.
  vector<int> iPartonMin;
  for (int i = legEnd[legMin]; i >= legBeg[legMin]; --i) { 
    int iNew = event.append( event[ iParton[i] ] ); 
    event[iNew].rotbst(MtoJRF);   
    iPartonMin.push_back( iNew ); 
  }
  vector<int> iPartonMid;
  for (int i = legEnd[legMid]; i >= legBeg[legMid]; --i) { 
    int iNew = event.append( event[ iParton[i] ] ); 
    event[iNew].rotbst(MtoJRF);   
    iPartonMid.push_back( iNew ); 
  }

  // Find final weighted sum of momenta on each of the two legs.
  double eWeight = 0.;
  pWTinJRF[legMin] = 0.; 
  for (int i = iPartonMin.size() - 1; i >= 0; --i) {
    pWTinJRF[legMin] += event[ iPartonMin[i] ].p() * exp(-eWeight);      
    eWeight += event[ iPartonMin[i] ].e() / eNormJunction; 
    if (eWeight > EJNWEIGHTMAX) break; 
  }
  eWeight = 0.;
  pWTinJRF[legMid] = 0.; 
  for (int i = iPartonMid.size() - 1; i >= 0; --i) {
    pWTinJRF[legMid] += event[ iPartonMid[i] ].p() * exp(-eWeight);      
    eWeight += event[ iPartonMid[i] ].e() / eNormJunction; 
    if (eWeight > EJNWEIGHTMAX) break; 
  }
    
  // Define fictitious opposing partons in JRF and store as string ends.
  Vec4 pOppose = pWTinJRF[legMin];   
  pOppose.flip3();
  int idOppose = (rndmPtr->flat() > 0.5) ? 2 : 1;
  if (event[ iPartonMin[0] ].col() > 0) idOppose = -idOppose;
  int iOppose = event.append( idOppose, 77, 0, 0, 0, 0, 0, 0,
    pOppose, 0.); 
  iPartonMin.push_back( iOppose);
  pOppose = pWTinJRF[legMid];   
  pOppose.flip3();
  idOppose = (rndmPtr->flat() > 0.5) ? 2 : 1;
  if (event[ iPartonMid[0] ].col() > 0) idOppose = -idOppose;
  iOppose = event.append( idOppose, 77, 0, 0, 0, 0, 0, 0,
    pOppose, 0.); 
  iPartonMid.push_back( iOppose);

  // Set up kinematics of string evolution in low-energy temporary systems.
  systemMin.setUp(iPartonMin, event);
  systemMid.setUp(iPartonMid, event);

  // Outer fallback loop, when too little energy left for third leg.
  int idMin = 0;
  int idMid = 0;
  Vec4 pDiquark;
  for ( int iTryOuter = 0; ; ++iTryOuter) {
  
    // Middle fallback loop, when much unused energy in leg remnants.
    double eLeftMin = 0.;
    double eLeftMid = 0.;
    for ( int iTryMiddle = 0; ; ++iTryMiddle) {  
    
      // Loop over the two lowest-energy legs.
      for (int legLoop = 0; legLoop < 2; ++ legLoop) { 
        int legNow = (legLoop == 0) ? legMin : legMid;

        // Read in properties specific to this leg.
        HelixStringSystem& systemNow = (legLoop == 0) ? systemMin : systemMid;
        int idPos = (legLoop == 0) ? event[ iPartonMin[0] ].id() 
          : event[ iPartonMid[0] ].id();
        idOppose = (legLoop == 0) ? event[ iPartonMin.back() ].id() 
          : event[ iPartonMid.back() ].id();
        double eInJRF = pInJRF[legNow].e();
        int statusHad = (legLoop == 0) ? 85 : 86; 

        // Inner fallback loop, when a diquark comes in to junction.
        double eUsed = 0.;
        for ( int iTryInner = 0; ; ++iTryInner) { 
          if (iTryInner > NTRYJNMATCH) {
            infoPtr->errorMsg("Error in HelixStringFragmentation::fragment" 
              "ToJunction: caught in junction flavour loop");
            event.popBack( iPartonMin.size() + iPartonMid.size() );
            return false;
          }

          // Set up two string ends, and begin fragmentation loop.
          setStartEnds(idPos, idOppose, systemNow);
          eUsed = 0.;
          int nHadrons = 0;
          bool noNegE = true;
          for ( ; ; ++nHadrons) {

           // Save hadron string composition for decays
	    HadronSystem hadron;

            // Construct trial hadron from positive end.
            posEnd.newHadron(hadronMode);
            Vec4 pHad = posEnd.kinematicsHadron(systemNow, hadron);

            // Negative energy signals failure in construction.
            if (pHad.e() < 0. ) { noNegE = false; break; }
  
            // Break if passed system midpoint ( = junction) in energy.
            if (eUsed + pHad.e() > eInJRF) break;

            // Else construct kinematics of the new hadron and store it.
            hadrons.append( posEnd.idHad, statusHad, iPos, iNeg, 
              0, 0, 0, 0, pHad, posEnd.mHad);

            // Update string end and remaining momentum.
            posEnd.update();
            eUsed += pHad.e();
          }

          // End of fragmentation loop. Inner loopback if ends on a diquark.
          if ( noNegE && abs(posEnd.flavOld.id) < 10 ) break; 
          hadrons.popBack(nHadrons);
        }

        // End of one-leg fragmentation. Store end quark and remnant energy.
        if (legNow == legMin) {
          idMin = posEnd.flavOld.id;
          eLeftMin = eInJRF - eUsed;
        } else {
          idMid = posEnd.flavOld.id; 
          eLeftMid = eInJRF - eUsed;
        }
      }

      // End of both-leg fragmentation. 
      // Middle loopback if too much energy left.
      double eTrial = eBothLeftJunction + rndmPtr->flat() * eMaxLeftJunction;
      if (iTryMiddle > NTRYJNMATCH 
        || ( min( eLeftMin, eLeftMid) < eBothLeftJunction
        && max( eLeftMin, eLeftMid) < eTrial ) ) break;
      hadrons.clear();
    }

    // Boost hadrons away from the JRF to the original frame.
    for (int i = 0; i < hadrons.size(); ++i) {
      hadrons[i].rotbst(MfromJRF);
      // Recalculate energy to compensate for numerical precision loss
      // in iterative calculation of MfromJRF.
      hadrons[i].e( hadrons[i].eCalc() );
      pJunctionHadrons += hadrons[i].p();
    }

    // Outer loopback if too little energy left in third leg.
    pDiquark = pInLeg[legMin] + pInLeg[legMid] - pJunctionHadrons; 
    double m2Left = m2( pInLeg[legMax], pDiquark);
    if (iTryOuter >  NTRYJNMATCH
      || m2Left > eMinLeftJunction * pInLeg[legMax].e() ) break;
    hadrons.clear();
    pJunctionHadrons = 0.;  
  }

  // Now found solution; no more loopback. Remove temporary parton copies.
  event.popBack( iPartonMin.size() + iPartonMid.size() ); 
  
  // Construct and store an effective diquark string end from the 
  // two remnant quark ends, for temporary usage. 
  int    idDiquark = flavSelPtr->makeDiquark( idMin, idMid);
  double mDiquark  = pDiquark.mCalc();
  int     iDiquark = event.append( idDiquark, 78, 0, 0, 0, 0, 0, 0,
    pDiquark, mDiquark);  

  // Find the partons on the last leg, again in reverse order.
  vector<int> iPartonMax;
  for (int i = legEnd[legMax]; i >= legBeg[legMax]; --i)
    iPartonMax.push_back( iParton[i] ); 
  iPartonMax.push_back( iDiquark ); 

  // Modify parton list to remaining leg + remnant of the first two.
  iParton = iPartonMax;
  
  // Done.
  return true;
}

//--------------------------------------------------------------------------

// Find the boost matrix to the rest frame of a junction,
// given the three respective endpoint four-momenta.

RotBstMatrix HelixStringFragmentation::junctionRestFrame(Vec4& p0, Vec4& p1, 
  Vec4& p2) {

  // Calculate masses and other invariants.
  Vec4 pSumJun  = p0 + p1 + p2;
  double sHat   = pSumJun.m2Calc();
  double pp[3][3];
  pp[0][0]      = p0.m2Calc();
  pp[1][1]      = p1.m2Calc();
  pp[2][2]      = p2.m2Calc();
  pp[0][1] = pp[1][0] = p0 * p1;  
  pp[0][2] = pp[2][0] = p0 * p2;  
  pp[1][2] = pp[2][1] = p1 * p2;  

  // Requirement (eiMax)_j = pi*pj/mj < (eiMax)_k = pi*pk/mk, used below, 
  // here rewritten as pi*pj * mk < pi*pk * mj and squared.
  double eMax01 = pow2(pp[0][1]) * pp[2][2];
  double eMax02 = pow2(pp[0][2]) * pp[1][1];
  double eMax12 = pow2(pp[1][2]) * pp[0][0];

  // Initially pick i to be the most massive parton. but allow other tries.
  int i = (pp[1][1] > pp[0][0]) ? 1 : 0;
  if (pp[2][2] > max(pp[0][0], pp[1][1])) i = 2; 
  int j, k;
  double ei     = 0.;
  double ej     = 0.;
  double ek     = 0.;
  for (int iTry = 0; iTry < 3; ++iTry) {

    // Pick j to give minimal eiMax, and k the third vector.
    if (i == 0) j = (eMax02 < eMax01) ? 2 : 1;
    else if (i == 1) j = (eMax12 < eMax01) ? 2 : 0;
    else j = (eMax12 < eMax02) ? 1 : 0; 
    k = 3 - i - j;

    // Alternative names according to i, j, k conventions.
    double m2i  = pp[i][i];
    double m2j  = pp[j][j];
    double m2k  = pp[k][k];
    double pipj = pp[i][j];
    double pipk = pp[i][k];
    double pjpk = pp[j][k];

    // Trivial to find new parton energies if all three partons are massless.
    if (m2i < M2MAXJRF) {
      ei        = sqrt( 2. * pipk * pipj / (3. * pjpk) );
      ej        = sqrt( 2. * pjpk * pipj / (3. * pipk) );
      ek        = sqrt( 2. * pipk * pjpk / (3. * pipj) );

    // Else find three-momentum range for parton i and values at extremes.
    } else { 
      // Minimum when i is at rest.
      double piMin = 0.; 
      double eiMin = sqrt(m2i);
      double ejMin = pipj / eiMin;
      double ekMin = pipk / eiMin;
      double pjMin = sqrtpos( ejMin*ejMin - m2j );   
      double pkMin = sqrtpos( ekMin*ekMin - m2k );  
      double fMin  = ejMin * ekMin + 0.5 * pjMin * pkMin - pjpk;
      // Maximum estimated when j + k is at rest, alternatively j at rest.
      double eiMax = (pipj + pipk) / sqrt(m2j + m2k + 2. * pjpk);
      if (m2j > M2MAXJRF) eiMax = min( eiMax, pipj / sqrt(m2j) );
      double piMax = sqrtpos( eiMax*eiMax - m2i ); 
      double temp  = eiMax*eiMax - 0.25 *piMax*piMax;
      double pjMax = (eiMax * sqrtpos( pipj*pipj - m2j * temp ) 
        - 0.5 * piMax * pipj) / temp;
      double pkMax = (eiMax * sqrtpos( pipk*pipk - m2k * temp ) 
        - 0.5 * piMax * pipk) / temp;
      double ejMax = sqrt(pjMax*pjMax + m2j);
      double ekMax = sqrt(pkMax*pkMax + m2k);
      double fMax  = ejMax * ekMax + 0.5 * pjMax * pkMax - pjpk;

      // If unexpected values at upper endpoint then pick another parton.
      if (fMax > 0.) {
        int iPrel = (i + 1)%3;
        if (pp[iPrel][iPrel] > M2MAXJRF) {i = iPrel; continue;}
        ++iTry;
        iPrel = (i + 2)%3;  
        if (iTry < 3 && pp[iPrel][iPrel] > M2MAXJRF) {i = iPrel; continue;}
      }

      // Start binary + linear search to find solution inside range.
      int iterMin = 0;
      int iterMax = 0;
      double pi   = 0.5 * (piMin + piMax);
      for (int iter = 0; iter < NTRYJRFEQ; ++iter) {
 
        // Derive momentum of other two partons and distance to root.
        ei = sqrt(pi*pi + m2i);
        temp = ei*ei - 0.25 * pi*pi;
        double pj = (ei * sqrtpos( pipj*pipj - m2j * temp )
          - 0.5 * pi * pipj) / temp;
        double pk = (ei * sqrtpos( pipk*pipk - m2k * temp )
          - 0.5 * pi * pipk) / temp;
        ej = sqrt(pj*pj + m2j);
        ek = sqrt(pk*pk + m2k);
        double fNow = ej * ek + 0.5 * pj * pk - pjpk;

        // Replace lower or upper bound by new value. 
        if (fNow > 0.) { ++iterMin; piMin = pi; fMin = fNow;}
        else {++iterMax; piMax = pi; fMax = fNow;}     
            
        // Pick next i momentum to explore, hopefully closer to root.
        if (2 * iter < NTRYJRFEQ 
          && (iterMin < 2 || iterMax < 2 || 4 * iter < NTRYJRFEQ))
          { pi = 0.5 * (piMin + piMax); continue;}  
        if (fMin < 0. || fMax > 0. || abs(fNow) < CONVJRFEQ * sHat) break;
        pi = piMin + (piMax - piMin) * fMin / (fMin - fMax);
      }

    // If arrived here then either succeeded or exhausted possibilities.
    } break;
  }

  // Now we know the energies in the junction rest frame.
  double eNew[3] = { 0., 0., 0.};  
  eNew[i] = ei;  
  eNew[j] = ej;  
  eNew[k] = ek;
  
  // Boost (copy of) partons to their rest frame.
  RotBstMatrix Mmove;  
  Vec4 p0cm = p0;  
  Vec4 p1cm = p1;  
  Vec4 p2cm = p2;  
  Mmove.bstback(pSumJun);
  p0cm.rotbst(Mmove);
  p1cm.rotbst(Mmove);
  p2cm.rotbst(Mmove); 

  // Construct difference vectors and the boost to junction rest frame.
  Vec4 pDir01      = p0cm / p0cm.e() - p1cm / p1cm.e();
  Vec4 pDir02      = p0cm / p0cm.e() - p2cm / p2cm.e();
  double pDiff01   = pDir01.pAbs2();
  double pDiff02   = pDir02.pAbs2();
  double pDiff0102 = dot3(pDir01, pDir02);  
  double eDiff01   = eNew[0] / p0cm.e() - eNew[1] / p1cm.e();
  double eDiff02   = eNew[0] / p0cm.e() - eNew[2] / p2cm.e();
  double denom     = pDiff01 * pDiff02 - pDiff0102*pDiff0102;
  double coef01    = (eDiff01 * pDiff02 - eDiff02 * pDiff0102) / denom;
  double coef02    = (eDiff02 * pDiff01 - eDiff01 * pDiff0102) / denom;
  Vec4 vJunction   = coef01 * pDir01 + coef02 * pDir02;
  vJunction.e( sqrt(1. + vJunction.pAbs2()) );

  // Add two boosts, giving final result.
  Mmove.bst(vJunction); 
  return Mmove;

}

//==========================================================================

  bool HelixStringFragmentation::helixDecay(int id) {
    
    if ( id == 113 || id == 213 || id == -213 ) return true;
    if ( id == 313 || id == 323 || id == -323 ) return true;
    
    return false;
  }

//==========================================================================

  bool HelixStringFragmentation::helixDecayResolved(int id, int& id1, int& id2, int& idnext) {
    
    int iFP = id1; int iFN = id2; 

    if (abs(iFP)>6 || abs(iFN)>6) return false;

    if (id == 113) {
      if (iFP==2 || iFP==-1) {
	id1=211; id2=-211; idnext = iFP==2 ? -1 : 2;
      } else {
	id1=-211; id2=211; idnext = iFP==1 ? -2 : 1;
      }
      return true;

    } else if (abs(id)==213) {
      idnext = rndmPtr->flat()>0.5 ? 1 : 2;
      if (iFP>0) idnext *=-1; 
      if (iFP+idnext==0) id1=111;
      else if (iFP+idnext>0) id1=211;
      else  id1=-211;
      if (iFN-idnext==0) id2=111;
      else if (iFN-idnext>0) id2=211;
      else  id2=-211;
      return true;
    } 
  
    if (id == 313 || abs(id)==323) {
      idnext = rndmPtr->flat()>0.5 ? 1 : 2;
      if (iFP>0) idnext *=-1; 
      if ( iFP==3 && idnext==-1) id1=311; 
      else if ( iFP==3 && idnext==-2)  id1=-321;
      else if ( iFP==-3 && idnext==2)  id1= 321;
      else if ( iFP==-3 && idnext==1)  id1= 311; 
      else if ( iFP+idnext==0) id1 = 111;
      else if ( iFP+idnext>0) id1 = 211;
      else if ( iFP+idnext<0) id1 =-211;
      
      if ( iFN==3 && idnext== 1) id2=311; 
      else if ( iFN==3 && idnext== 2)  id2=-321;
      else if ( iFN==-3 && idnext==-2)  id2= 321;
      else if ( iFN==-3 && idnext==-1)  id2= 311; 
      else if ( iFN-idnext==0) id2 = 111;
      else if ( iFN-idnext>0) id2 = 211;
      else if ( iFN-idnext<0) id2 =-211;
      return true;
    }  
   
    return false;
  }

//==========================================================================

  void HelixStringFragmentation::decayHelix() {
  
    
    for ( int i=0; i<resonances.size();i++ ) {
    
      int id = hadrons[resonances[i].first].id();
      HadronSystem had = resonances[i].second;
 
      int id1=had.flavPos; int id2=had.flavNeg; int idnext=0;

      bool dec=false;
      if ( helixDecayResolved(id,id1,id2,idnext)  )
	  dec = splitHelix(had,id1,id2,resonances[i].first,idnext);

    }

  }
//==========================================================================

  bool HelixStringFragmentation::splitHelix(HadronSystem& hsys, int id1, int id2, int imother, int iflav) {

    double mPos = particleDataPtr->mass(id1);
    double mNeg = particleDataPtr->mass(id2);

    int chiral = hsys.phasePos < hsys.phaseNeg ? 1 : -1; 

    Vec4 pTot = hsys.getMomentum(system);
     
    vector< pair<int, double> > phPos;
 
    StringRegion regp;
    regp.setUp(system.regionLowPos(hsys.iPP).pPos,system.regionLowNeg(hsys.iPN).pNeg,true);
    StringRegion regn;
    regn.setUp(system.regionLowPos(hsys.iNP).pPos,system.regionLowNeg(hsys.iNN).pNeg,true);

    // find solutions
    vector<pair<HadronSystem,HadronSystem>  >solutions;
    vector<pair<HadronSystem,HadronSystem>  >apsols;
    Vec4 pP = regp.pHad(0.,0.,hsys.xP,hsys.yP);
    for (unsigned int jl=0; jl<hsys.frPos.size(); jl++)  {
      int il = hsys.frPos[jl].first; 
      Vec4 pM = regn.pHad(0.,0.,0.,0.);
      vector< pair<int, double> > phNeg;
      phPos.push_back(pair<int,double> (il, 0.));
      //for (unsigned int jr=0; jr< hsys.frNeg.size(); jr++)  {
      for (int jr=hsys.frNeg.size()-1; jr >-1; jr--)  {
	int ir = hsys.frNeg[jr].first; 
	phNeg.push_back(pair<int,double> (ir, 0.));
	if (hsys.frPos[jl].second!=0 && hsys.frNeg[jr].second!=0 
	    && il>=hsys.iPP && il<=hsys.iNP && ir<=hsys.iPN && ir>=hsys.iNN ) {
          // collect known momenta from pos and neg side, project on current region 
          StringRegion reg;
          Vec4 pPlus = hsys.frPos[jl].second*system.regionLowPos(il).pPos;
          Vec4 pMins = hsys.frNeg[jr].second*system.regionLowNeg(ir).pNeg;
          reg.setUp(pPlus,pMins, true);
          // side momenta   
          Vec4 pPos = pP+pM ; Vec4 pNeg = pTot - pPos - pPlus - pMins ;
          // projections
	  reg.project( pTot);
	  double pxRem   = reg.px();
	  double pyRem   = reg.py();
	  double xPosRem = reg.xPos();
	  double xNegRem = reg.xNeg();
	  reg.project( pPos);
	  double pxPos   = reg.px();
	  double pyPos   = reg.py();
	  double xPosPos = reg.xPos();
	  double xNegPos = reg.xNeg();
	  reg.project( pNeg);
	  double pxNeg   = reg.px();
	  double pyNeg   = reg.py();
	  double xPosNeg = reg.xPos();
	  double xNegNeg = reg.xNeg();

          //double mat2 = pxPos*pxPos + pyPos*pyPos + mPos*mPos;
          //double mbt2 = pxNeg*pxNeg + pyNeg*pyNeg + mNeg*mNeg;
	  double mat2 =  mPos*mPos;
          double mbt2 =  mNeg*mNeg;

          // longitudinal rest frame
          double er = sqrt( pTot.m2Calc()+pxRem*pxRem+pyRem*pyRem );
          double ea = 0.5*(er*er+mat2-mbt2)/er;
	  double dd = pow2(er*er-mat2-mbt2) - 4*mat2*mbt2;
     
          if (xPosRem*xNegRem>0 && reg.w2 > 0 && ea > 0 && dd > 0 ) {
	    int jd = (mPos!=mNeg ) ? 1 : 2;
	    for (int isign=+1; isign>-2; isign=isign-2*jd) {
	      //double lambda = sqrt(dd);       
	      //double probReverse = 1. / (1. + exp( min( EXPMAX, bLund * lambda) ) ); 
	      double pl = sqrt(dd)/2/er;
	      //int isign = +1;
	      //if (probReverse > rndmPtr->flat()) isign=-1;
	      pl *= isign;  
	      double xp = (ea+pl)/er; double xm = (ea-pl)/er;
	      Vec4 hPos = reg.pHad( xp*xPosRem, xm*xNegRem, pxPos,pyPos);
	      double mhp = hPos.mCalc();
	      Vec4 hNeg = pTot - hPos;
	      double mhn = hNeg.mCalc();
	      // recalculate helix phase   
	      //cout << "current phase:"<<posEnd.phaseOld+posEnd.phaseDifference(phPos,phNeg,system)<< endl;
	      phPos.back()=pair<int,double> (il, (xp*xPosRem-xPosPos)*hsys.frPos[jl].second );
	      phNeg.back()=pair<int,double> (ir, (xm*xNegRem-xNegPos)*hsys.frNeg[jr].second );
	      double phaseHad = posEnd.phaseDifference(phPos,phNeg,system); 
	      double phaseSplit = hsys.phasePos + chiral*phaseHad; 
	      //cout << "helix phase split:"<< phaseSplit<<endl;
              double rad = radius+sigma_radius*rndmPtr->gauss();
	      double pxN = rad*cos(phaseSplit-chiral*0.5*M_PI);
	      double pyN = rad*sin(phaseSplit-chiral*0.5*M_PI);
	      mat2 = (pxPos+pxN)*(pxPos+pxN) + (pyPos+pyN)*(pyPos+pyN) + mPos*mPos;
	      mbt2 = (pxNeg-pxN)*(pxNeg-pxN) + (pyNeg-pyN)*(pyNeg-pyN) + mNeg*mNeg;
	      double ean = 0.5*(er*er+mat2-mbt2)/er;
	      double ddn = pow2(er*er-mat2-mbt2) - 4*mat2*mbt2;
	      int iter=0;
	      double phCurr = phaseSplit; 
	      //cout << "recalculated:" << ean <<","<<ddn<< endl;
	      //cout<<"man,mbn,ern,ean,ddn:"<<mat2<<","<<mbt2<<","<<er<<","<<ean<<","<<ddn<<endl;
	      //cout << "mhp,mhn:"<< mhp<<","<<mhn<< endl;          
	      
	      //while (ean > 0 && ddn > 0 && iter<5 && mhp>0 && mhn>0) {
	      while (ean > 0 && ddn > 0 && iter<5) {
		double pln = isign*sqrt(ddn)/2/er;
		double xpn = (ean+pln)/er; double xmn = (ean-pln)/er;
                //cout <<"xpn,xmn:"<<xpn<<","<<xmn<<","<<xpn*xmn<<","<<xPosRem*xNegRem<< endl;
		hPos = reg.pHad( xp*xPosRem, xm*xNegRem, pxPos+pxN,pyPos+pyN);
		double xps = xpn*xPosRem-xPosPos;
		double xns = xmn*xNegRem-xNegPos;
		
		phPos.back()= pair<int,double> (il, xps*hsys.frPos[jl].second );
		phNeg.back()= pair<int,double> (ir, xns*hsys.frNeg[jr].second );
		phaseHad = posEnd.phaseDifference(phPos,phNeg,system); 
		phaseSplit = hsys.phasePos + chiral*phaseHad;

		if ( fabs(phaseSplit-phCurr)<0.01 && (phaseSplit-hsys.phasePos)*(hsys.phaseNeg-phaseSplit)>0 ) {
		  bool good_sol =  xps>=0. && xps+xPosPos<=(xPosRem-xPosNeg) 
		    && xns>=0. && xns+xNegPos<=(xNegRem-xNegNeg) ;
		  
		  HadronSystem dauPos; HadronSystem dauNeg;
		  for (int ip = 0; ip<hsys.frPos.size(); ip++) {
		    int jp = hsys.frPos[ip].first;
		    if (ip<jl) dauPos.frPos.push_back(pair<int,double>(jp,hsys.frPos[ip].second));
		    else if (ip==jl) {
		      dauPos.frPos.push_back(pair<int,double>(jp,(xpn*xPosRem-xPosPos)*hsys.frPos[ip].second));
		      dauNeg.frPos.push_back(pair<int,double>(jp,(1.-xpn*xPosRem+xPosPos)*hsys.frPos[ip].second));
		    } else  dauNeg.frPos.push_back(pair<int,double>(jp,hsys.frPos[ip].second));
		  } 
		  for (int in=hsys.frNeg.size()-1; in >-1; in--)  {
		    //for (int in = 0; in<hsys.frNeg.size(); in++) {
		    int jn = hsys.frNeg[in].first;

		    if (in>jr) dauPos.frNeg.push_back(pair<int,double>(jn,hsys.frNeg[in].second));
		    else if (in==jr) {
		      dauPos.frNeg.push_back(pair<int,double>(jn,(xmn*xNegRem-xNegPos)*hsys.frNeg[in].second));
		      dauNeg.frNeg.push_back(pair<int,double>(jn,(1.-xmn*xNegRem+xNegPos)*hsys.frNeg[in].second));
		    } else  dauNeg.frNeg.push_back(pair<int,double>(jn,hsys.frNeg[in].second));
		  } 
		  dauPos.setEnds(hsys.iPP,hsys.iPN,hsys.xP,hsys.yP,il,ir,pxN,pyN);
		  dauNeg.setEnds(il,ir,-pxN,-pyN, hsys.iNP,hsys.iNN,hsys.xN,hsys.yN);
                  dauPos.setPhases(hsys.phasePos,phaseSplit);
                  dauNeg.setPhases(phaseSplit,hsys.phaseNeg);
                  dauPos.setFlavours(hsys.flavPos,iflav);
                  dauNeg.setFlavours(-iflav,hsys.flavNeg);
		  Vec4 dPos= dauPos.getMomentum(system); 
		  Vec4 dNeg= dauNeg.getMomentum(system);
                  bool verify_decay = fabs(dPos.mCalc()-mPos)<1.e-06 && fabs(dNeg.mCalc()-mNeg)<1.e-06;
                  if (verify_decay) {
		    if (good_sol) solutions.push_back(pair<HadronSystem,HadronSystem>(dauPos,dauNeg));
		    else apsols.push_back(pair<HadronSystem,HadronSystem>(dauPos,dauNeg));
		    iter=5;
		  }
		}
		pxN = rad*cos(phaseSplit-chiral*0.5*M_PI);
		pyN = rad*sin(phaseSplit-chiral*0.5*M_PI);
		mat2 = (pxPos+pxN)*(pxPos+pxN) + (pyPos+pyN)*(pyPos+pyN) + mPos*mPos;
		mbt2 = (pxNeg-pxN)*(pxNeg-pxN) + (pyNeg-pyN)*(pyNeg-pyN) + mNeg*mNeg;
		ean = 0.5*(er*er+mat2-mbt2)/er;
		ddn = pow2(er*er-mat2-mbt2) - 4*mat2*mbt2;
		phCurr = phaseSplit;
		iter++;
	      }
	    }
          }
	}
	pM += hsys.frNeg[jr].second*system.regionLowNeg(ir).pNeg;        
        phNeg.back()=pair<int,double> (ir, hsys.frNeg[jr].second );
      }
      pP += hsys.frPos[jl].second*system.regionLowPos(il).pPos;
      phPos.back()=pair<int,double> (il, hsys.frPos[jl].second );
    }


    // multiple solutions ?
    if (!solutions.size()  && !apsols.size() ) {
      //cout << "decayHelix fails" << endl; 
      return false;
 
    } else if (!solutions.size()) {
    
      double sum = 0.; vector<double> gs;
      for (int is=0;is < apsols.size(); is++ ) {
	// estimate the Gamma factor
	double gamma = 2*apsols[is].first.getMomentumNeg(system)*apsols[is].second.getMomentumPos(system);
        gs.push_back( exp(-bLund*gamma) );
        sum += gs[is];
      }
      double w = sum * rndmPtr->flat();
      sum = 0.; int is=-1;
      while (sum < w ) {is++;sum += gs[is];}
      if (is<0 || is >= apsols.size()) is = int(apsols.size()*rndmPtr->flat());
      //cout <<" selected solution : " << is << endl;
      // Add produced particles to the event record.
      if (imother==0) {
	hadrons.append( id1, 83, posEnd.iEnd, negEnd.iEnd, 
			0, 0, 0, 0, apsols[is].first.getMomentum(system), mPos);
        if ( doHelixDecay && helixDecay(id1) ) resonances.push_back(pair<int,HadronSystem>(hadrons.size()-1,apsols[is].first));
	hadrons.append( id2, 84, posEnd.iEnd, negEnd.iEnd, 
			0, 0, 0, 0, apsols[is].second.getMomentum(system) , mNeg);
        if ( doHelixDecay && helixDecay(id2) ) resonances.push_back(pair<int,HadronSystem>(hadrons.size()-1,apsols[is].second));
        //cout <<"final split solution +:"<<id1<<":"<<apsols[is].first.flavPos<<","<<apsols[is].first.flavNeg<< endl;
        //cout <<"final split solution -:"<<id2<<":"<<apsols[is].second.flavPos<<","<<apsols[is].second.flavNeg<< endl;
      } else {
	hadrons.append( id1, 91, imother, imother,
			0, 0, 0, 0, apsols[is].first.getMomentum(system) , mPos);
        if ( doHelixDecay && helixDecay(id1) ) resonances.push_back(pair<int,HadronSystem>(hadrons.size()-1,apsols[is].first));
	hadrons.append( id2, 91, imother, imother, 
			0, 0, 0, 0, apsols[is].second.getMomentum(system), mNeg);
        if ( doHelixDecay && helixDecay(id2) ) resonances.push_back(pair<int,HadronSystem>(hadrons.size()-1,apsols[is].second));
        // add daughter pointers
        hadrons[imother].daughter1(hadrons.size()-2);
        hadrons[imother].daughter2(hadrons.size()-1);
      }
      return true;

    } else {
     
      double sum = 0.; vector<double> gs;
      for (int is=0;is < solutions.size(); is++ ) {
	// estimate the Gamma factor
	double gamma = 2*solutions[is].first.getMomentumNeg(system)*solutions[is].second.getMomentumPos(system);
        gs.push_back( exp(-bLund*gamma) );
        sum += gs[is];
      }
      double w = sum * rndmPtr->flat();
      sum = 0.; int is=-1;
      while (sum < w ) {is++;sum += gs[is];}
      if (is<0 || is >= solutions.size()) is = int(solutions.size()*rndmPtr->flat());
      //cout <<" selected solution : " << is << endl;
      // Add produced particles to the event record.
      if (imother==0) {
	hadrons.append( id1, 83, posEnd.iEnd, negEnd.iEnd, 
			0, 0, 0, 0, solutions[is].first.getMomentum(system), mPos);
        if ( doHelixDecay && helixDecay(id1) ) resonances.push_back(pair<int,HadronSystem>(hadrons.size()-1,solutions[is].first));
	hadrons.append( id2, 84, posEnd.iEnd, negEnd.iEnd, 
			0, 0, 0, 0, solutions[is].second.getMomentum(system) , mNeg);
        if ( doHelixDecay && helixDecay(id2) ) resonances.push_back(pair<int,HadronSystem>(hadrons.size()-1,solutions[is].second));
        //cout <<"final split solution +:"<<id1<<":"<<solutions[is].first.flavPos<<","<<solutions[is].first.flavNeg<< endl;
        //cout <<"final split solution -:"<<id2<<":"<<solutions[is].second.flavPos<<","<<solutions[is].second.flavNeg<< endl;
      } else {
	hadrons.append( id1, 91, imother, imother,
			0, 0, 0, 0, solutions[is].first.getMomentum(system) , mPos);
        if ( doHelixDecay && helixDecay(id1) ) resonances.push_back(pair<int,HadronSystem>(hadrons.size()-1,solutions[is].first));
	hadrons.append( id2, 91, imother, imother, 
			0, 0, 0, 0, solutions[is].second.getMomentum(system), mNeg);
        if ( doHelixDecay && helixDecay(id2) ) resonances.push_back(pair<int,HadronSystem>(hadrons.size()-1,solutions[is].second));
        // add daughter pointers
        hadrons[imother].daughter1(hadrons.size()-2);
        hadrons[imother].daughter2(hadrons.size()-1);
      }
      return true;
       
    }
    return false; 

  }
    
//==========================================================================

} // end namespace Pythia8
